"""Test picorun classes."""

import picorun.errors


def test_api_error() -> None:
    """Test ApiError."""
    error = picorun.errors.ApiError("message", 404)
    assert str(error) == "Error 404: message"
    assert error.status_code == 404



"""Test picorun classes."""

import pytest

import picorun
import picorun.errors


def mock_response(
    status_code: int = 200,
    headers: dict[str, str] | None = None,
    text: str | None = None,
    json: dict[str, str] | None = None,
) -> "MockResponse":  # noqa: F821 class defined in this function
    """Mock requests.Response."""

    class MockResponse:
        def __init__(self) -> None:
            self.status_code = status_code
            self.headers = headers or {}
            self.text = text
            self.json = lambda: json

    return MockResponse()


def test_api_request_args() -> None:
    """Test ApiRequestArgs."""
    args = picorun.picorun.ApiRequestArgs(
        path={"path": "value"},
        query={"query": "value"},
        payload={"payload": "value"},
        headers={"headers": "value"},
    )
    assert args.path == {"path": "value"}
    assert args.query == {"query": "value"}
    assert args.json == {"payload": "value"}
    assert args.headers == {"headers": "value"}
    assert args.to_kwargs() == {
        "headers": {"headers": "value"},
        "json": {"payload": "value"},
        "query": {"query": "value"},
    }


def test_api_request_empty_args() -> None:
    """Test ApiRequestArgs with empty arguments."""
    args = picorun.picorun.ApiRequestArgs()
    assert args.path == {}
    assert args.query == {}
    assert args.json == {}
    assert args.headers == {}
    assert args.to_kwargs() == {
        "headers": {},
        "json": {},
        "query": {},
    }


def test_api_response() -> None:
    """Test ApiResponse."""
    response = mock_response(
        status_code=200,
        headers={"Content-Type": "text/plain"},
        text="body",
    )
    api_response = picorun.picorun.ApiResponse(response)
    assert api_response.response == response
    assert api_response.status_code == 200
    assert api_response.headers == {"Content-Type": "text/plain"}
    assert api_response.body == "body"
    assert api_response.asdict() == {
        "statusCode": 200,
        "headers": {"Content-Type": "text/plain"},
        "body": "body",
    }


def test_api_response_no_content_type() -> None:
    """Test ApiResponse with no content type header."""
    response = mock_response(status_code=200, text="")
    api_response = picorun.picorun.ApiResponse(response)
    assert api_response.body == ""


def test_api_response_json() -> None:
    """Test ApiResponse with JSON body."""
    response = mock_response(
        status_code=200,
        headers={"Content-Type": "application/json"},
        json={"key": "value"},
    )
    api_response = picorun.picorun.ApiResponse(response)
    assert api_response.body == {"key": "value"}
    assert api_response.asdict() == {
        "statusCode": 200,
        "headers": {"Content-Type": "application/json"},
        "body": {"key": "value"},
    }


@pytest.mark.parametrize(
    ("code", "exp"),
    [
        (400, picorun.errors.Http400Error),
        (401, picorun.errors.Http401Error),
        (403, picorun.errors.Http403Error),
        (404, picorun.errors.Http404Error),
        (405, picorun.errors.Http405Error),
        (418, picorun.errors.Http418Error),
        (429, picorun.errors.Http429Error),
        (500, picorun.errors.Http500Error),
        (501, picorun.errors.Http501Error),
        (502, picorun.errors.Http502Error),
        (451, picorun.errors.ApiError),
    ],
)
def test_api_response_error_errors(code: int, exp: picorun.errors.ApiError) -> None:
    """Test ApiResponse with 400 Bad Request Error."""
    response = mock_response(
        status_code=code,
        text="",
    )
    api_response = picorun.picorun.ApiResponse(response)
    with pytest.raises(exp):
        api_response.raise_for_status()



"""PicoRun Errors."""


class ApiError(Exception):
    """Exception raised when an API call fails."""

    def __init__(self, message: str, status_code: int) -> None:
        """Construct an ApiError."""
        super().__init__(f"Error {status_code}: {message}")
        self.status_code = status_code


class Http400Error(ApiError):
    """HTTP 400 Bad Request Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 400 Error."""
        super().__init__(message, 400)


class Http401Error(ApiError):
    """HTTP 401 Unauthorised Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 401 Error."""
        super().__init__(message, 401)


class Http403Error(ApiError):
    """HTTP 403 Forbidden Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 403 Error."""
        super().__init__(message, 403)


class Http404Error(ApiError):
    """HTTP 404 Not Found Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 404 Error."""
        super().__init__(message, 404)


class Http405Error(ApiError):
    """HTTP 405 Method Not Allowed Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 405 Error."""
        super().__init__(message, 405)


class Http418Error(ApiError):
    """HTTP 418 Teapot Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 418 Error."""
        super().__init__(message, 418)


class Http429Error(ApiError):
    """HTTP 429 Too Many Requests Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 429 Error."""
        super().__init__(message, 429)


class Http500Error(ApiError):
    """HTTP 500 Internal Server Error Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 500 Error."""
        super().__init__(message, 500)


class Http501Error(ApiError):
    """HTTP 501 Not Implemented Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 501 Error."""
        super().__init__(message, 501)


class Http502Error(ApiError):
    """HTTP 502 Bad Gateway Error."""

    def __init__(self, message: str) -> None:
        """Construct a HTTP 502 Error."""
        super().__init__(message, 502)



"""PicoRun runtime classes."""

from typing import Any, ClassVar

from requests import Response

import picorun.errors


class ApiRequestArgs:
    """Arguments for making an API call with requests."""

    def __init__(
        self,
        path: dict[str, Any] | None = None,
        query: dict[str, Any] | None = None,
        payload: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
    ) -> None:
        """Construct an object with the arguments for the API call."""
        self.path = path if isinstance(path, dict) else {}
        self.query = query if isinstance(query, dict) else {}
        self.json = payload if isinstance(payload, dict) else {}
        self.headers = headers if isinstance(headers, dict) else {}

    def to_kwargs(self) -> dict[str, Any]:
        """Convert the object to a dictionary of keyword arguments for requests."""
        output = {}
        for property in ["headers", "json", "query"]:
            output[property] = getattr(self, property)
        return output


class ApiResponse:
    """API response."""

    _error_mapping: ClassVar[dict[int : picorun.errors.ApiError]] = {
        400: picorun.errors.Http400Error,
        401: picorun.errors.Http401Error,
        403: picorun.errors.Http403Error,
        404: picorun.errors.Http404Error,
        405: picorun.errors.Http405Error,
        418: picorun.errors.Http418Error,
        429: picorun.errors.Http429Error,
        500: picorun.errors.Http500Error,
        501: picorun.errors.Http501Error,
        502: picorun.errors.Http502Error,
    }

    def __init__(self, response: Response) -> None:
        """Construct an ApiResponse."""
        self.response = response
        self.status_code = response.status_code
        self.headers = response.headers
        self.body = response.text
        if (
            "Content-Type" in response.headers
            and "application/json" in response.headers["Content-Type"]
        ):
            self.body = response.json()

    def asdict(self) -> dict[str, Any]:
        """Convert the object to a dictionary."""
        return {
            "statusCode": self.status_code,
            "headers": dict(self.headers),
            "body": self.body,
        }

    def raise_for_status(self) -> None:
        """
        Raise an exception if an error response is received.

        This method is inspired by the Requests' method of the same name. The difference
        is PicoRun raises classed errors for many common HTTP error codes. This makes it
        easier to handle the errors in Step Functions.
        """
        if self.status_code in self._error_mapping:
            exp = self._error_mapping[self.status_code]
            raise exp(self.body)

        if self.status_code >= 400:
            raise picorun.errors.ApiError(self.body, self.status_code)
